"""Display subsystem"""
from rsfsup.common import Subsystem, validate


class Display(Subsystem, kind="Display"):
    """Display subsystem

    Attributes:
        instr (Fsup)
    """

    _formats = {"SING": "FULL SCREEN", "SPL": "SPLIT SCREEN"}
    _windows = {"1": "SCREEN A", "2": "SCREEN B"}
    _couplings = {
        "NONE": "NONE",
        "RLEVEL": "REF LEVEL",
        "CF_B": "CENTER B = MARKER A",
        "CF_A": "CENTER A = MARKER B",
    }

    def __init__(self, instr):
        super().__init__(instr)
        self._coupling = "NONE"
        self._visa.write(f"INSTRUMENT:COUPLE {self._coupling}")
        self._psave_holdoff = 0
        self._visa.write(f"DISPLAY:PSAVE:STATE OFF")

    @property
    def format(self):
        """value (str): {FULL SCREEN, SPLIT SCREEN}"""
        result = self._visa.query("DISPLAY:FORMAT?")
        return Display._formats.get(result, result)

    @format.setter
    @validate
    def format(self, value):
        formats = dict((v, k) for k, v in Display._formats.items())
        value = formats.get(value, value)
        self._visa.write(f"DISPLAY:FORMAT {value}")

    @property
    def screen(self):
        """value (str): {SCREEN A, SCREEN B}"""
        result = self._visa.query("DISPLAY:ACTIVE?")
        return Display._windows.get(result, result)

    @screen.setter
    @validate
    def screen(self, value):
        windows = dict((v, k) for k, v in Display._windows.items())
        try:
            value = windows[value]
        except KeyError:
            raise ValueError(f"{value} not a valid screen")
        self._visa.write(f"DISPLAY:WINDOW{value}:SELECT")

    def set_title(self, text=""):
        """text (str): active screen title up to 20 characters"""
        num = self._visa.query("DISPLAY:ACTIVE?")
        if text == "":
            self._visa.write(f"DISPLAY:WINDOW{num}:TEXT:STATE OFF")
        else:
            self._visa.write(f"DISPLAY:WINDOW{num}:TEXT:DATA '{text[:20]}'")
            self._visa.write(f"DISPLAY:WINDOW{num}:TEXT:STATE ON")

    @property
    def coupling(self):
        """value (str): {NONE, REF LEVEL, CENTER B = MARKER A, CENTER A = MARKER A}"""
        return self._coupling

    @coupling.setter
    @validate
    def coupling(self, value):
        couplings = dict((v, k) for k, v in Display._couplings.items())
        try:
            value = couplings[value]
        except KeyError:
            raise ValueError(f"{value} not a valid coupling")
        self._visa.write(f"INSTRUMENT:COUPLE {value}")

    @property
    def pwr_save(self):
        """value (int): screen-saver timeout in minutes (1 to 60), 0 disables pwr save"""
        return self._psave_holdoff

    @pwr_save.setter
    @validate
    def pwr_save(self, value):
        self._psave_holdoff = value
        if value:
            self._visa.write(f"DISPLAY:PSAVE:HOLDOFF {value}")
            self._visa.write("DISPLAY:PSAVE:STATE ON")
        else:
            self._visa.write("DISPLAY:PSAVE:STATE OFF")



"""Common definitions"""
from collections import namedtuple
import functools

IDN = namedtuple("IDN", ["manufacturer", "model", "serial_number", "firmware_version"])

FREQUENCY_UNITS = {0: "Hz", 3: "kHz", 6: "MHz", 9: "GHz"}

TIME_UNITS = {0: "s", 3: "ms", 6: "µs", 9: "ns"}

BOOLEAN = {"0": "OFF", "1": "ON"}


def get_idn(visa):
    """Get IDN"""
    idn = visa.query("*IDN?").strip(":\n").split(",")
    return IDN(*idn)


class CommandError(Exception):
    """Raised when CME bit of SESR is set"""


def validate(func):
    """Read the Command Error bit (CME) of the Event-Status Register (ESR)"""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        # pylint: disable=protected-access
        cme = 32
        self = args[0]
        self._visa.write("*CLS")
        self._visa.write("SYSTEM:ERROR:CLEAR:ALL")
        original_ese = int(self._visa.query("*ESE?"))
        self._visa.write(f"*ESE {cme}")
        result = func(*args, **kwargs)
        if bool(int(self._visa.query("*ESR?")) & cme):
            err_id, err_msg = self._visa.query("SYSTEM:ERROR?").split(",")
            err_id = int(err_id)
            err_msg = err_msg.strip('"\r')
            raise CommandError(err_msg)
        self._visa.write(f"*ESE {original_ese}")
        return result

    return wrapper


class RSBase:
    """Instrument base class

    Attributes:
        visa (pyvisa.resources.Resource): pyvisa resource
    """

    def __init__(self, visa):
        self._visa = visa

    def __setattr__(self, name, value):
        if hasattr(self, name) and isinstance(getattr(self, name), RSBase):
            raise AttributeError(f"can't set '{name}'")
        if hasattr(self.__class__, name):
            prop = getattr(self.__class__, name)
            try:
                prop.fset(self, value)
            except TypeError:
                raise AttributeError(f"can't set '{name}'") from None
        else:
            self.__dict__[name] = value

    def __repr__(self):
        raise NotImplementedError

    def __dir__(self):
        inst_attr = list(filter(lambda k: not k.startswith("_"), self.__dict__.keys()))
        cls_attr = list(filter(lambda k: not k.startswith("_"), dir(self.__class__)))
        return inst_attr + cls_attr

    def _screen(self):
        """Get the active screen"""
        return self._visa.query("DISPLAY:ACTIVE?")


class Subsystem(RSBase):
    """Base class for a specific measurement function

    Attributes:
        instr (Fsup)
    """

    _kind = "Instrument"

    def __init__(self, instr):
        super().__init__(instr._visa)
        self._instr = instr

    def __init_subclass__(cls, kind):
        cls._kind = kind

    def __get__(self, instance, owner=None):
        return self

    def __repr__(self):
        return f"<{self._instr.model} {self._kind}>"


def scale_frequency(value):
    """value (float): convert value to quantity"""
    exp = 0
    scaled = value
    while scaled >= 1.0 and exp < 9:
        exp += 3
        scaled = value / 10 ** exp
    unit = FREQUENCY_UNITS[exp]
    return f"{scaled} {unit}"


def scale_time(value):
    """value (float): convert value to quantity"""
    exp = 0
    scaled = value
    while scaled < 1.0 and exp < 12:
        exp += 3
        scaled = value * 10 ** exp
    unit = TIME_UNITS[exp]
    return f"{scaled} {unit}"



"""RSFSUP - an interface to the Rohde&Schwarz FSUP Signal Source Analyzer

Basic usage to plot the spectrum:
>>> from rsfsup import CommChannel
>>> with CommChannel("<ip address>") as fsup:
...     spectrum = fsup.read()
...
>>> import matplotlib.pyplot as plt
>>> plt.plot(*spectrum)
[<matplotlib.lines.Line2D object at ...>]
>>> plt.show()
"""
import pyvisa
from unyt import define_unit, matplotlib_support
from rsfsup.common import get_idn
from rsfsup.fsup import Fsup
from rsfsup.version import __version__

__all__ = ["CommChannel"]

define_unit("dBm", (1.0, "dB"))
matplotlib_support()
matplotlib_support.label_style = "/"


class CommChannel:
    """Connect to a R&S FSUP using VISA

    Attributes:
        address (str): instrument's TCPIP address or host name

    Returns:
        CommChannel or Fsup
    """

    def __init__(self, address):
        self._address = address
        self._rm = pyvisa.ResourceManager()
        self._visa = self._rm.open_resource(f"TCPIP::{address}")
        self._visa.read_termination = "\n"

    def __enter__(self):
        self._visa.write("*CLS")
        idn = get_idn(self._visa)
        if idn.manufacturer != "Rohde&Schwarz" and not idn.model.startswith("FSUP"):
            raise ValueError(f"Device at {self._address} is a not a R&S FSUP")
        self._visa.write("SYSTEM:DISPLAY:UPDATE ON")
        self._visa.write("SYSTEM:DISPLAY:FPANEL ON")
        return Fsup(self._visa)

    def __exit__(self, exc_type, exc_value, exc_tb):
        self._visa.write("SYSTEM:DISPLAY:FPANEL ON")
        self._visa.write("SYSTEM:KLOCK OFF")
        self._visa.close()
        self._rm.close()

    def get_instrument(self):
        """Return the Fsup instrument object"""
        return self.__enter__()

    def close(self):
        """Close the CommChannel"""
        self.__exit__(None, None, None)


def __dir__():
    return ["CommChannel", "__version__"]



"""Fsup class"""
from datetime import datetime
import numpy as np
from unyt import unyt_quantity, unyt_array
from rsfsup.common import RSBase, get_idn, validate
from rsfsup.display import Display
from rsfsup.frequency import Frequency
from rsfsup.amplitude import Amplitude
from rsfsup.bandwidth import Bandwidth
from rsfsup.sweep import Sweep
from rsfsup.markers import Marker, DeltaMarker
from rsfsup.traces import Trace


class Fsup(RSBase):
    """Fsup class

    Attributes:
        visa (pyvisa.resources.Resource): pyvisa resource
    """

    _modes = {"SAN": "SPECTRUM", "PNO": "SSA"}
    _options = {
        "B4": "Low-aging OCXO",
        "B10": "External Generator Control",
        "B18": "Removable Hard Disk",
        "B19": "Second Hard Disk",
        "B21": "LO/IF Ports for External Mixers",
        "B23": "20 dB Preamplifier",
        "B25": "Electronic Attenuator",
        "B28": "Trigger Port",
        "B60": "Low Phase Noise",
        "B61": "Correlation Extension",
        "K5": "GSM/EDGE",
        "K8": "Bluetooth",
        "K9": "Power Sensor Measurements",
        "K30": "Noise Figure and Gain Measurements",
        "K70": "Vector Signal Analysis",
    }

    def __init__(self, visa):
        super().__init__(visa)
        self._visa.write("*CLS")
        self._visa.write("SYSTEM:ERROR:CLEAR:ALL")
        self._visa.write("ESE 255")
        self._visa.write("FORMAT ASCII")
        self._idn = get_idn(visa)
        self.display = Display(self)
        self.frequency = Frequency(self)
        self.amplitude = Amplitude(self)
        self.bandwidth = Bandwidth(self)
        self.sweep = Sweep(self)
        self._markers = [Marker(self, num) for num in range(1, 5)]
        self._deltamarkers = [DeltaMarker(self, num) for num in range(1, 5)]
        self._traces = [Trace(self, 1)]

    @property
    def model(self):
        """(str): the model number"""
        return self._idn.model

    @property
    def serial_number(self):
        """(str): the serial number"""
        return self._idn.serial_number

    @property
    def firmware_version(self):
        """(str): the firmware version"""
        return self._idn.firmware_version

    @property
    def mode(self):
        """(str): {SPECTRUM, SSA}"""
        result = self._visa.query("INSTRUMENT?")
        return Fsup._modes.get(result, result)

    @mode.setter
    @validate
    def mode(self, value):
        modes = dict((v, k) for k, v in Fsup._modes.items())
        try:
            value = modes[value]
        except KeyError:
            raise ValueError(f"{value} not a valid mode") from None
        self._visa.write(f"INSTRUMENT:SELECT {value}")

    def reset(self):
        """Reset the instrument to the factory default settings"""
        self._visa.write("*RST")

    def set_clock(self):
        """Set the date and time to local machine's time"""
        self._visa.write(f"SYSTEM:DATE {datetime.now().strftime('%Y,%m,%d')}")
        self._visa.write(f"SYSTEM:TIME {datetime.now().strftime('%H,%M,%S')}")

    def lock_frontpanel(self):
        """Lock the front panel"""
        self._visa.write("SYSTEM:DISPLAY:FPANEL OFF")
        self._visa.write("SYSTEM:KLOCK ON")

    def unlock_frontpanel(self):
        """Unlock the front panel"""
        self._visa.write("SYSTEM:KLOCK OFF")
        self._visa.write("SYSTEM:DISPLAY:FPANEL ON")

    @property
    def options(self):
        """List the installed options"""
        codes = self._visa.query("*OPT?").split(",")
        installed_codes = [code for code in codes if not code.startswith("0")]
        return [Fsup._options.get(code, code) for code in installed_codes]

    @property
    def impedance(self):
        """value (int): RF input impedance {50, 75} Ω"""
        return int(self._visa.query("INPUT:IMPEDANCE?"))

    @impedance.setter
    @validate
    def impedance(self, value):
        self._visa.write(f"INPUT:IMPEDANCE {value}")

    def turnoff_markers(self):
        """Turn off all markers in the active screen"""
        for marker in self._markers + self._deltamarkers:
            marker.state = "OFF"

    def reference_fixed(self, state="ON"):
        """state (str): {ON, OFF}
            Turn on markers 1 and 2, if necessary, and set markers 2 to 4 as delta
            markers fixed to marker 1
        """
        self.enable_marker(1)
        self.enable_marker(2, as_delta=True)
        self._visa.write(f"CALC{self._screen()}:DELT2:FUNC:FIX {state}")

    def enable_marker(self, num, as_delta=False):
        """num (int): marker/deltamarker number {1, 2, 3, 4}
        as_delta (bool): True if deltamarker"""
        i = num - 1
        marker = self._markers[i]
        deltamarker = self._deltamarkers[i]
        # pylint: disable=protected-access
        if as_delta:
            deltamarker.state = "ON"
            try:
                setattr(self, deltamarker._name, deltamarker)
            except AttributeError:
                pass
            marker.state = "OFF"
            try:
                delattr(self, marker._name)
            except AttributeError:
                pass
        else:
            deltamarker.state = "OFF"
            try:
                delattr(self, deltamarker._name)
            except AttributeError:
                pass
            marker.state = "ON"
            try:
                setattr(self, marker._name, marker)
            except AttributeError:
                pass

    def clear_traces(self):
        """Clear all traces"""
        self._visa.write(f"DISP:WIND{self._screen()}:TRACE:CLEAR")

    def read(self, trace=1):
        """Read trace data and return tuple (X, Y)

        Parameters:
            trace (int): {1, 2, 3}
        """
        num = self.sweep.points
        start_str = self.frequency.start
        value, unit = start_str.split(" ")
        start = unyt_quantity(float(value), unit)
        stop_str = self.frequency.stop
        value, unit = stop_str.split(" ")
        stop = unyt_quantity(float(value), unit)
        x = np.linspace(start, stop, num=num, endpoint=True)
        x.name = "$f$"
        i = trace - 1
        data = self._traces[i].data
        unit = self._traces[i].y_unit
        y = unyt_array(data, unit)
        y.name = "$A$"
        return (x, y)

    def __dir__(self):
        for marker in self._markers + self._deltamarkers:
            if marker.state == "OFF":
                try:
                    delattr(self, marker._name)
                except AttributeError:
                    pass
        return super().__dir__()

    def __repr__(self):
        return f"<R&S {self.model} at {self._visa.resource_name}>"



"""Frequency subsystem"""
from rsfsup.common import Subsystem, scale_frequency, validate


class Frequency(Subsystem, kind="FREQ"):
    """Frequency subsystem

    Attributes:
        instr (Fsup)
    """

    @property
    def center(self):
        """value (int or str):
            0 to f_max in hertz
            string such as '1 GHz'
            keyword such as 'UP' or 'DOWN'
        """
        value = float(self._visa.query(f"SENSE{self._screen()}:FREQUENCY:CENTER?"))
        return scale_frequency(value)

    @center.setter
    @validate
    def center(self, value):
        self._visa.write(f"SENSE{self._screen()}:FREQUENCY:CENTER {value}")

    @property
    def span(self):
        """value (int or str): 0 to f_max in hertz or string such as '100 MHz' or 'FULL'"""
        value = float(self._visa.query(f"SENSE{self._screen()}:FREQUENCY:SPAN?"))
        return scale_frequency(value)

    @span.setter
    @validate
    def span(self, value):
        if value == "FULL":
            self._visa.write(f"SENSE{self._screen()}:FREQUENCY:SPAN:FULL")
        else:
            self._visa.write(f"SENSE{self._screen()}:FREQUENCY:SPAN {value}")

    @property
    def start(self):
        """value (int or str): 0 to f_max in hertz or string such as '20 MHz'"""
        value = float(self._visa.query(f"SENSE{self._screen()}:FREQUENCY:START?"))
        return scale_frequency(value)

    @start.setter
    @validate
    def start(self, value):
        self._visa.write(f"SENSE{self._screen()}:FREQUENCY:START {value}")

    @property
    def stop(self):
        """value (int or str): 0 to f_max in hertz or string such as '20 MHz'"""
        value = float(self._visa.query(f"SENSE{self._screen()}:FREQUENCY:STOP?"))
        return scale_frequency(value)

    @stop.setter
    @validate
    def stop(self, value):
        self._visa.write(f"SENSE{self._screen()}:FREQUENCY:STOP {value}")

    @property
    def cf_step(self):
        """value (int or string): {f Hz, SPAN, CENTER, MARKER}"""
        value = float(self._visa.query(f"SENSE{self._screen()}:FREQ:CENT:STEP?"))
        return scale_frequency(value)

    @cf_step.setter
    @validate
    def cf_step(self, value):
        if value == "SPAN":
            self._visa.write(f"SENSE{self._screen()}:FREQ:CENT:STEP:LINK SPAN")
        elif value == "CENTER":
            cf = self.center
            self._visa.write(f"SENSE{self._screen()}:FREQ:CENT:STEP:LINK OFF")
            self._visa.write(f"SENSE{self._screen()}:FREQ:CENT:STEP {cf}")
        elif value == "MARKER":
            self._visa.write(f"CALCULATE{self._screen()}:MARKER:FUNCTION:CSTEP")
        else:
            self._visa.write(f"SENSE{self._screen()}:FREQ:CENT:STEP:LINK OFF")
            self._visa.write(f"SENSE{self._screen()}:FREQ:CENT:STEP {value}")
